/*
 * Trivial arpwatch. Support Linux /proc/net/arp only. Do not detect flip-flop.
 * Support call script if detect new or changes ARP for IP.
 *
 * Copyright (c) 1999-2014 may safely be consumed by a BSD or GPL license.
 * Written by:   Vladimir Oleynik <dzo@simtreas.ru>
 *
 * bb_*error*():
 *      Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *      Hacked by me for use syslog in daemon mode
 * xmalloc() bb_xstrdup():
 *      Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * last_char_is():
 *      Copyright (C) 2001 Larry Doolittle, <ldoolitt@recycle.lbl.gov>
 *
 *
 * other functions by me
 */

#define DEFAULT_WAIT_SEC 5
#define DEFAULT_WAIT_SEC_STR "5"
#define PID_DIR "/var/run"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <errno.h>
#include <stdarg.h>
#include <syslog.h>
#include <signal.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <resolv.h>


#define MAX_TOKENS 3


static long safe_strtol(const char *arg)
{
	char *endptr;
	long value = value;     /* make gcc happy */

	if(arg == NULL || *arg == '\0')
		goto eret;

	errno = 0;
	value = strtol(arg, &endptr, 0);
	if (*endptr != '\0') {
eret:
		errno = EINVAL;
	}
	return value;
}

#ifndef ARP_DAT
#define ARP_DAT "/usr/local/etc/arpwatch/arpwatch.dat"
#endif
#define ARP_DAT_FORMAT "IP address       HW address            Timestamp"
#define ARP_DAT_WORDS     3
#define DAT_IP_WORD_NUM   1
#define DAT_ARP_WORD_NUM  2
#define DAT_TIME_WORD_NUM 3
typedef struct ARP_DAT_LINE
{
	struct ARP_DAT_LINE *link;
	unsigned int narp[6];
	char *tokens[MAX_TOKENS];
	struct in_addr addr;
	long ntime;
	struct ARP_DAT_LINE *new_station;
} arp_dat_line_t;
static int arp_dat_nums[MAX_TOKENS+1] = { DAT_IP_WORD_NUM, DAT_ARP_WORD_NUM, DAT_TIME_WORD_NUM, ARP_DAT_WORDS };

static arp_dat_line_t *top_arp_dat;

static int callback_arp_dat(void *line)
{
	arp_dat_line_t *cur = line;
	if (inet_aton(cur->tokens[0], &(cur->addr)) == 0)
		return -1;
	if(sscanf(cur->tokens[1], "%2x:%2x:%2x:%2x:%2x:%2x",
			&(cur->narp[0]), &(cur->narp[1]), &(cur->narp[2]),
			&(cur->narp[3]), &(cur->narp[4]), &(cur->narp[5])) != 6)
		return -2;
	cur->ntime = safe_strtol(cur->tokens[2]);
	if(errno != 0)
		return -3;
	cur->new_station = NULL;
	return 1;
}

#ifndef PROC_ARP
#define PROC_ARP "/proc/net/arp"
#endif
#define PROC_ARP_FORMAT "IP address       HW type     Flags       HW address            Mask     Device"
#define PROC_ARP_WORDS    6
#define IP_WORD_NUM       1
#define ARP_WORD_NUM      4
#define DEV_WORD_NUM      6
typedef struct PROC_ARP_LINE
{
	struct PROC_ARP_LINE *link;
	unsigned int narp[6];
	char *tokens[MAX_TOKENS];
	struct in_addr addr;
} proc_arp_line_t;
static int proc_arp_nums[MAX_TOKENS+1] = { IP_WORD_NUM, ARP_WORD_NUM, DEV_WORD_NUM, PROC_ARP_WORDS };

static proc_arp_line_t *top_arp;

static char **devs;
static int ndevs;

static int callback_proc_arp(void *line)
{
	proc_arp_line_t *cur = line;

	if (inet_aton(cur->tokens[0], &(cur->addr)) == 0)
		return -1;
	if(strcmp(cur->tokens[1], "00:00:00:00:00:00") == 0)
		return 0;       /* skip incomplete */
	if(sscanf(cur->tokens[1], "%2x:%2x:%2x:%2x:%2x:%2x",
			&(cur->narp[0]), &(cur->narp[1]), &(cur->narp[2]),
			&(cur->narp[3]), &(cur->narp[4]), &(cur->narp[5])) != 6)
		return -2;
	if(devs != NULL) {
		int i;

		for(i = 0; i < ndevs; i++) {
			if(strcmp(devs[i], cur->tokens[2]) == 0)
				return 1;
		}
		return 0;       /* skip unneed device */
	}
	return 1;
}

#ifndef ETHER_DAT
#define ETHER_DAT "/usr/local/etc/arpwatch/ethercodes.dat"
#endif
#define ETHER_DAT_FORMAT "HW address\tFirm"
#define ETHER_DAT_WORDS        2
#define CODE_WORD_NUM          1
#define VENDOR_WORD_NUM       -2
typedef struct ETHER_DAT_LINE
{
	struct ETHER_DAT_LINE *link;
	unsigned int narp[6];
	char *tokens[MAX_TOKENS];
	char *vendor;
} codes_line_t;
static int ether_dat_nums[MAX_TOKENS+1] = { CODE_WORD_NUM, VENDOR_WORD_NUM, 0, ETHER_DAT_WORDS };

static codes_line_t *top_codes;


static const char usage_msg[] ="[-p " PROC_ARP "] [-d " ARP_DAT "] [-c " ETHER_DAT
				"] [-w(sleep) " DEFAULT_WAIT_SEC_STR "] "
				"[-i comma_separated_devices_list] [-f(oreground)] "
				"[-s script] [-l logfile] [-P " PID_DIR "]";

typedef struct LINE
{
	struct LINE *link;
	unsigned int narp[6];
	char *tokens[MAX_TOKENS];
} line_t;


/*
 */

static int log_stderr = 1, log_syslog;
static const char *bb_applet_name;

static void bb_error_msg_and_die(const char *s, ...) __attribute__ ((noreturn, format (printf, 1, 2)));
static void bb_perror_msg_and_die(const char *s, ...) __attribute__ ((noreturn, format (printf, 1, 2)));
static void bb_show_usage (void) __attribute__ ((__noreturn__));

static void bb_verror_msg(const char *s, va_list p)
{
	if(log_syslog) {
		va_list copy;

		va_copy(copy, p);
		vsyslog(LOG_ERR, s, p);
		va_end(copy);
	}

	if(log_stderr) {
		fprintf(stderr, "%s: ", bb_applet_name);
		vfprintf(stderr, s, p);
		putc('\n', stderr);
	}
	va_end(p);
}

static void bb_vperror_msg(int err, const char *s, va_list p)
{
	if(log_syslog) {
		va_list copy;
		char *se = alloca(strlen(s) + 5);       /* ": %m\0" */

		sprintf(se, "%s: %%m", s);
		va_copy(copy, p);
		errno = err;
		vsyslog(LOG_ERR, se, p);
		va_end(copy);
	}

	if(log_stderr) {
		fprintf(stderr, "%s: ", bb_applet_name);
		vfprintf(stderr, s, p);
		fprintf(stderr, ": %s\n", strerror(err));
	}
	va_end(p);
}

static void bb_perror_msg_and_die(const char *s, ...)
{
	int err = errno;
	va_list p;

	va_start(p, s);
	bb_vperror_msg(err,  s, p);
	exit(2);
}

static void bb_error_msg_and_die(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	bb_verror_msg(s, p);
	exit(1);
}

#if 0
static void bb_error_msg(const char *s, ...) __attribute__ ((format (printf, 1, 2)));
static void bb_error_msg(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	bb_verror_msg(s, p);
}
#endif

static const char bb_msg_memory_exhausted[]="memory exhausted";

static void *xmalloc(size_t size)
{
	void *ptr = malloc(size);
	if (ptr == NULL && size != 0)
		bb_error_msg_and_die(bb_msg_memory_exhausted);
	return ptr;
}

static char * bb_xstrdup (const char *s)
{
	char *t;

	if (s == NULL)
		return NULL;

	t = strdup (s);

	if (t == NULL)
		bb_error_msg_and_die(bb_msg_memory_exhausted);

	return t;
}


static void bb_show_usage (void)
{
	fprintf(stderr, "Usage %s %s\n", bb_applet_name, usage_msg);
	exit(3);
}

/* by me codepoet code, but simplified special for this program */
/* simplified:
   1.   GNU : "+" "-" "::"
   2.         
'int',          
  .
   3. applet_opts    "ab:..."   
\0,       \0 :
   -  '-'  ,         
 ,   
   -  ,     
:

	int flags;

	char *b_str = default_b_opt_value;
	flags = xgetopt(argc, argv, "ab:\0", &b_str);
        -a -b b_value  -


	char *b_str;
	flags = xgetopt(argc, argv, "ab:\0b", &b_str);
        -a   
 -b b_value  .     
b_str .

	char *b_str = default_b_opt_value;
	flags = xgetopt(argc, argv, "ab:\0-", &b_str);
	      -a -b b_value

	char *c_str;
	flags = xgetopt(argc, argv, "ab:c:\0-c", &b_str, &c_str);
     ,       -a 
-b b_value   -c c_value     

	 flags:
	if((flags & 1)) {      -a }
	if((flags & 2)) {      -b b_value }
	if((flags & 4)) {      -c c_value }
*/
static int xgetopt(int argc, char **argv, const char *applet_opts, ...)
{
	const char *p;
	const char ***a;
	int c;
	int opt;
	int ret = 0;
	va_list vl;

	p = strrchr(argv[0], '/');
	if(p != NULL)
		bb_applet_name = p + 1;
	else
		bb_applet_name = argv[0];

	/* need applet_opts != NULL and applet_opts[0] != '\0' */
	a = alloca(strlen(applet_opts) * sizeof(char **));

	va_start (vl, applet_opts);
	for(c = 0, p = applet_opts; *p != '\0'; c++) {
		if(*++p == ':') {
			a[c] = va_arg (vl, const char **);
			p++;
		} else
			a[c] = NULL;
	}
	va_end(vl);

	while((opt=getopt(argc, argv, applet_opts)) != -1) {
		for(c = 0, p = applet_opts; *p != '\0'; c++) {
			if(*p == opt) {
				if(a[c] != NULL)
					*(a[c]) = optarg;
				if((unsigned)c >= (sizeof(int)*8))
					c = sizeof(int)*8 - 1;
				ret |= (1 << c);
				break;
			}
			if(*++p == ':')
				p++;
		}
		if(*p == '\0')
			bb_show_usage ();
	}

	argv += optind;
	/* test need options */
	for(p = applet_opts; *p != '\0'; p++)
		;
	if(*++p == '\0')
		return ret;     /* do not check need args */
	if(*p == '-') {
		p++;
		if(*argv == NULL)
			bb_show_usage ();       /* need arguments */
	} else if(*argv != NULL)                /* do not need arguments */
		bb_show_usage ();

	{
		const char *o = p;
		while(*o != '\0') {
			for(c = 0, p = applet_opts; *p != '\0'; c++) {
				if(p[1] == ':') {
					if(*o == *p && *a[c] == NULL)
						bb_show_usage ();
					p++;
				}
				p++;
			}
			o++;
		}
	}
	return ret;
}

static void *
xrealloc (void *p, size_t size)
{
  p = realloc (p, size);
  if (p == NULL)
	bb_error_msg_and_die(bb_msg_memory_exhausted);
  return p;
}


static int callback_ether_dat(void *line)
{
	codes_line_t *cur = line;

	if(sscanf(cur->tokens[0], "%2x:%2x:%2x",
			&(cur->narp[0]), &(cur->narp[1]), &(cur->narp[2])) != 3)
		return -1;
	cur->vendor = bb_xstrdup(cur->tokens[1]);
	return 1;
}


static void parser (const char *filename, int need_skip_line, size_t sizeof_struct,
			void *ptop, const char *format, int nums[MAX_TOKENS+1], int (*callback)(void *line))
{
  size_t in_buff_size = 128;            /*    */
  char *input_buff = xmalloc(in_buff_size);
  int readed;
  char *line;
  char delim;
  line_t *cur;
  int num_token;
  int num_line = 0;
  size_t token_size;
  FILE *in;
  struct LINE **link = (struct LINE **)ptop;
  int i;

  in = fopen(filename, "r");
  if(in == NULL)
	bb_perror_msg_and_die("can not open %s for read", filename);

  readed = 0;
  do {
	line = fgets (input_buff + readed, in_buff_size - readed, in);
	if(line == NULL) {
		/* EOF   / */
		if(ferror(in))
			bb_perror_msg_and_die("IO error while read %s", filename);
		if(readed == 0) {
			break;
		} else {
			/*       */
			readed = -readed;
		}
	} else {
		line = strchr(input_buff, '\n');
		if(line == NULL) {
			/*   in_buff_size */
			readed = in_buff_size - 1;
			input_buff = xrealloc(input_buff, in_buff_size *= 2);
			continue;
		} else {
			if((line - 1) != input_buff && line[-1] == '\\') {
				/*    \\\n */
				num_line++;
				readed = strlen(input_buff) - 2;
				continue;
			}
			readed = 0;
		}
	}
	num_line++;
	if(need_skip_line > 0) {
		need_skip_line--;
		continue;
	}

	cur = xmalloc (sizeof_struct);
	for(i = 0; i < MAX_TOKENS; i++)
		cur->tokens[i] = NULL;
	num_token = 0;
	line = input_buff;
	delim = '\0';

	while (*line) {
	  token_size = strcspn (line, " \t\n");
	  delim = line[token_size];
	  line[token_size] = '\0';
	  if (token_size) {
		num_token++;
		for(i = 0; i < MAX_TOKENS; i++) {
		  if(num_token == nums[i]) {
			cur->tokens[i] = line;
			break;
		  } else if(num_token == -nums[i]) {
			char *e;
			/* special parse: "WORD strings..." */
			line[token_size] = delim;
			e = strchr(line, '\n');
			if(e != NULL)
				*e = '\0';
			token_size = strlen(line);
			delim = '\0';
			cur->tokens[i] = line;
			break;
		  }
		}
	  }
	  /*          */
	  line += token_size + (delim != '\0');
	}

	if (num_token != nums[MAX_TOKENS]) {
  err:
	  bb_error_msg_and_die("%s have a strange format in line %d, we can parse as:\n%s", filename, num_line, format);
	}
	i = (*callback)(cur);
	if(i < 0)
		goto err;
	if(i == 0) {
		free(cur);
	} else {
		cur->link = *link;
		*link = cur;
	}

  } while (readed >= 0);

  free(input_buff);
  fclose(in);
}

static const char *find_vendor(void *v)
{
	codes_line_t *codes;
	line_t *cur = v;

	for(codes = top_codes; codes != NULL; codes = codes->link) {
		if(cur->narp[0] == codes->narp[0] &&
			cur->narp[1] == codes->narp[1] &&
			cur->narp[2] == codes->narp[2])
				return codes->vendor;
	}
	return "<unknow>";
}

static void *free_list(void *list)
{
  line_t *l, *top = list;

  for(l = top; l != NULL; ) {
	top = top->link;
	free (l);
	l = top;
  }
  return NULL;
}

static const char *pidfilename;

static void unlink_pidfile(void)
{
	if(pidfilename != NULL) {
		unlink(pidfilename);
		pidfilename = NULL;
	}
}

static void sigterm_handler(int sig)
{
	exit(sig);
}

static volatile int reload;
static void reload_handler(int sig)
{
	reload = sig;
}

static char *bb_xasprintf(const char *format, ...) __attribute__ ((format (printf, 1, 2)));
static char *bb_xasprintf(const char *format, ...)
{
	va_list p;
	int r;
	char *string_ptr;

	va_start(p, format);
	r = vasprintf(&string_ptr, format, p);
	va_end(p);

	if (r < 0)
		bb_perror_msg_and_die("bb_xasprintf");
	return string_ptr;
}


static char *cwd;

static char *xgetcwd_const_dir (void)
{
  static const int PATH_INCR = 32;
  char *ret;
  unsigned path_max = 1026-PATH_INCR;

  if(cwd != NULL)
	return cwd;

  do {
      path_max += PATH_INCR;
      free(cwd);
      cwd = xmalloc (path_max);
  } while ((ret = getcwd (cwd, path_max)) == NULL && errno == ERANGE);

  if (ret == NULL)
      bb_perror_msg_and_die("getcwd()");

  return cwd;
}

static char * last_char_is(const char *s, int c)
{
	char *sret = (char *)s;
	if (sret) {
		sret = strrchr(sret, c);
		if(sret != NULL && *(sret+1) != 0)
			sret = NULL;
	}
	return sret;
}

/* by me bugreports */
static char *concat_path_file(const char *path, const char *filename)
{
	char *lc;

	if (!path)
		path = "";
	lc = last_char_is(path, '/');
	while (*filename == '/')
		filename++;
	return bb_xasprintf("%s%s%s", path, (lc==NULL ? "/" : ""), filename);
}


/* hack by me for const chdir and some bugreports */
static char *bb_simplify_path_const_cwd(const char *path)
{
	char *s, *start, *p;

	s = path[0] == '/' ? bb_xstrdup(path) : concat_path_file(xgetcwd_const_dir(), path);
	p = start = s;

	do {
		if (*p == '/') {
			if (*s == '/') {        /* skip duplicate (or initial) slash */
				continue;
			} else if (*s == '.') {
				if (s[1] == '/' || s[1] == 0) { /* remove extra '.' */
					continue;
				} else if ((s[1] == '.') && (s[2] == '/' || s[2] == 0)) {
					++s;
					if (p > start) {
						while (*--p != '/');    /* omit previous dir */
					}
					continue;
				}
			}
		}
		*++p = *s;
	} while (*++s);

	if ((p == start) || (*p != '/')) {      /* not a trailing slash */
		++p;                                    /* so keep last character */
	}
	*p = 0;

	return start;
}

static void xstat(const char *name, struct stat *pst)
{
	if(stat(name, pst))
		bb_perror_msg_and_die("can not stat %s", name);
}

static char *make_fullpath_and_tstisdir(const char *d)
{
	struct stat st;
	char *r = bb_simplify_path_const_cwd(d);

	xstat(d, &st);
	if(S_ISDIR(st.st_mode) == 0)
		bb_error_msg_and_die("%s is not a directory", d);
	return r;
}

static void set_signal_intr(int signo, sighandler_t h)
{
	struct sigaction act;

	memset(&act, 0, sizeof(act));
	act.sa_handler = h;
	while(sigaction(signo, &act, NULL) < 0 && errno == EINTR)
		;
}

static void xdaemon(const char *pid_dir, const char *logfile)
{
	FILE *fl;
	char *p;

	/* create_pidfile */
	p = make_fullpath_and_tstisdir(pid_dir);
	pidfilename = bb_xasprintf("%s/%s.pid", p, bb_applet_name);
	free(p);
	fl = fopen(pidfilename, "w");
	if(fl == NULL)
		bb_perror_msg_and_die("can not create %s file", pidfilename);
	set_signal_intr(SIGTERM, sigterm_handler);
	set_signal_intr(SIGQUIT, sigterm_handler);
	set_signal_intr(SIGINT, sigterm_handler);
	set_signal_intr(SIGHUP, sigterm_handler);
	atexit(unlink_pidfile);

	log_syslog = 1; /* syslog+stderr. stderr have set /dev/null before have detect fork() errors */
	if(daemon(0, 0))
		bb_perror_msg_and_die("daemon()");

	if(logfile != NULL) {
		/* stdout (as /dev/null after daemon()) -> logfile */
		if(freopen(logfile, "a", stdout) == NULL)
			bb_perror_msg_and_die("can not create/open %s", logfile);
		setvbuf(stdout, (char *) NULL, _IOLBF, 0);
		if(freopen(logfile, "a", stderr) == NULL)
			bb_perror_msg_and_die("can not create/open %s", logfile);
		setvbuf(stderr, (char *) NULL, _IOLBF, 0);
		log_syslog = 0;
	} else
		log_stderr = 0; /* syslog only */

	fprintf(fl, "%d\n", getpid());
	fclose(fl);
}

static char *tst_relative_script(char *script)
{
	if(script != NULL) {
		struct stat st;
		char *s = bb_simplify_path_const_cwd(script);

		xstat(s, &st);
		if(S_ISREG(st.st_mode) == 0)
			bb_error_msg_and_die("%s is not a file", script);
		return s;
	}
	return script;
}

static void run_script(char *script)
{
	int k_pid;
	volatile int vfork_exec_errno;
	char *kargv[2];

	kargv[0] = (char *)script;
	kargv[1] = NULL;

	vfork_exec_errno = 0;
	k_pid = vfork();

	if(k_pid < 0)
		bb_perror_msg_and_die("vfork() for kill");
	if(k_pid == 0) {
		execvp(script, kargv);
		vfork_exec_errno = errno;
		_exit(-1);
	}
	if(vfork_exec_errno) {
		errno = vfork_exec_errno;
		bb_perror_msg_and_die("can not start %s", script);
	}
	while(waitpid(k_pid, NULL, 0) != k_pid)
		;
}

static sigset_t bb_sigmask(unsigned int sigs_bits)
{
	unsigned int b = 2;
	int signo = 1;
	sigset_t m;

	sigemptyset(&m);
	while(b) {
		if((sigs_bits & b) != 0)
			sigaddset(&m, signo);
		signo++;
		b <<= 1;
	}
	return m;
}

static sigset_t bb_sigblock(unsigned int sigs_bits)
{
	sigset_t m;

	m = bb_sigmask(sigs_bits);
	sigprocmask(SIG_BLOCK, &m, NULL);
	return m;
}

static void bb_sigunblock(sigset_t m)
{
	sigprocmask(SIG_UNBLOCK, &m, NULL);
}

static void print_host_ip(proc_arp_line_t *proc, time_t t, int is_not_first)
{
	int32_t options;
	struct hostent *hp;

	if(is_not_first == 0)
		printf("Time: %s", ctime(&t));
	printf("%s", inet_ntoa(proc->addr));
	options = _res.options;
	_res.options |= RES_AAONLY;
	_res.options &= ~(RES_DEFNAMES | RES_DNSRCH);
	hp = gethostbyaddr(&(proc->addr), sizeof(proc->addr), AF_INET);
	_res.options = options;
	if (hp != NULL)
		printf("(%s)", hp->h_name);
}

static char *show_arp(void *line)
{
	line_t *l = line;
	static char arp_buff[6*3];

	sprintf(arp_buff, "%02x:%02x:%02x:%02x:%02x:%02x",
		l->narp[0], l->narp[1], l->narp[2],
		l->narp[3], l->narp[4], l->narp[5]);
	return arp_buff;
}


int
main (int argc, char **argv)
{
  int r;
  const char *log = NULL;
  const char *proc_arp = PROC_ARP;
  const char *arp_dat = ARP_DAT;
  const char *ether_dat = ETHER_DAT;
  const char *pid_dir = PID_DIR;
  const char *ws = NULL;
  int wait_sec = DEFAULT_WAIT_SEC;
  char *script = NULL;
  int foreground;
  proc_arp_line_t *proc;
  arp_dat_line_t *dat;
  time_t t;
  char *devlist = NULL;
  int out;
  int need_change;

  r = xgetopt(argc, argv, "fl:p:d:c:i:w:P:s:\0X",
	&log, &proc_arp, &arp_dat, &ether_dat, &devlist, &ws, &pid_dir, &script);
  foreground = r & 1;

  if(!foreground) {
	/* make absolute paths */
	if(log != NULL)
		log = bb_simplify_path_const_cwd(log);
	proc_arp = bb_simplify_path_const_cwd(proc_arp);
	arp_dat = bb_simplify_path_const_cwd(arp_dat);
	ether_dat = bb_simplify_path_const_cwd(ether_dat);
	signal(SIGUSR1, reload_handler);
  }
  script = tst_relative_script(script);

  if(devlist != NULL) {
	char *d;

	r = 0;
	for(d = devlist; *d != '\0'; d++)
		if(*d == ',')
			r++;
	devs = xmalloc(sizeof(char *) * (r + 1));
	d = bb_xstrdup(devlist);
	while (*d != '\0') {
	  size_t token_size = strcspn (d, ",");
	  char delim = d[token_size];
	  if (token_size) {
		d[token_size] = '\0';
		devs[ndevs++] = d;
	  } else {
		bb_error_msg_and_die("strange the -i `%s' argument", devlist);
	  }
	  d += token_size + (delim != '\0');
	}
  }

  if(ws != NULL && ((wait_sec = safe_strtol(ws)) <= 0 || errno != 0))
	bb_error_msg_and_die("strange the -w `%s' argument", ws);

  out = foreground != 0 || log != NULL;

  parser (arp_dat, 0, sizeof(arp_dat_line_t), &top_arp_dat, ARP_DAT_FORMAT, arp_dat_nums, callback_arp_dat);

  if(!foreground)
	xdaemon(pid_dir, log);

  for(;;) {
    if(top_codes == NULL || reload != 0) {
	codes_line_t *codes;

	for(codes = top_codes; codes != NULL; codes = codes->link)
		free(codes->vendor);
	top_codes = free_list(top_codes);
	parser (ether_dat, 0, sizeof(codes_line_t), &top_codes, ETHER_DAT_FORMAT, ether_dat_nums, callback_ether_dat);
	reload = 0;
    }

    need_change = 0;
    time(&t);
    parser (proc_arp, 1, sizeof(proc_arp_line_t), &top_arp, PROC_ARP_FORMAT, proc_arp_nums, callback_proc_arp);

    for(proc = top_arp; proc != NULL; proc = proc->link) {
	for(dat = top_arp_dat; dat != NULL; dat = dat->link) {
		if(proc->addr.s_addr == dat->addr.s_addr) {
			if(memcmp(proc->narp, dat->narp, sizeof(proc->narp)) != 0) {
				if(out) {
					const char *pv, *nv;

					print_host_ip(proc, t, need_change);
					printf(" changed arp to %s ", show_arp(proc));
					printf("from %s\n", show_arp(dat));
					pv = find_vendor(dat);
					nv = find_vendor(proc);
					if(pv == nv) {
						printf("Same vendor is %s\n", pv);
					} else {
						printf("Previous vendor is %s\n", pv);
						printf("New vendor is %s\n", nv);
					}
					printf("Previous timestamp %s", ctime(&dat->ntime));
				}
				memcpy(dat->narp, proc->narp, sizeof(dat->narp));
				dat->ntime = t;
				need_change |= 1;
			}
			break;
		}
	}
	if(dat == NULL) {
		if(out) {
			print_host_ip(proc, t, need_change);
			printf(" is new station with ARP=%s\nVendor is %s\n", show_arp(proc), find_vendor(proc));
		}
		dat = xmalloc (sizeof(arp_dat_line_t));
		dat->addr = proc->addr;
		memcpy(dat->narp, proc->narp, sizeof(dat->narp));
		dat->ntime = t;
		dat->link = top_arp_dat;
		top_arp_dat = dat;
		dat->new_station = dat;
		need_change |= 2;       /* add only flag */
	}
    }

    if(need_change) {
	sigset_t m;
	FILE *dat_out;

	m = bb_sigblock((1<<SIGTERM)|(1<<SIGQUIT)|(1<<SIGINT)|(1<<SIGHUP));
	dat_out = fopen(arp_dat, need_change == 2 ? "a" : "w");
	if(dat_out == NULL)
		bb_perror_msg_and_die("can not overwrite %s", arp_dat);

	for(dat = top_arp_dat; dat != NULL; dat = dat->link) {
		if(need_change != 2 || dat->new_station != NULL) {
			fprintf(dat_out, "%s\t%s\t%ld\n", inet_ntoa(dat->addr), show_arp(dat), dat->ntime);
			dat->new_station = NULL;
		}
	}
	fclose(dat_out);
	bb_sigunblock(m);

	if(script != NULL) {
		if(out)
			printf("run %s\n", script);
		run_script(script);
	}
	if(out)
		putchar('\n');
    }

    if(foreground && ws == NULL)
	return 0;

    top_arp = free_list(top_arp);
    sleep(wait_sec);
  }
}
